
__all__ = [
    "ES_kernel_name",
    "ES_CODE"    
]
ES_kernel_name = "ES_kernel"

ES_CODE = '''
#define i blockIdx.x
#define j blockIdx.y
#define tid_x threadIdx.x
#define tid_y threadIdx.y
#define bDm_x blockDim.x
#define bDm_y blockDim.y
#define WARP_SIZE 32

__forceinline__ __device__ int rowMajor_2D(int r, int c, int R, int C){
    return c + C * r;
}

extern "C"{
__global__ void ES_kernel(float * C, int const * const grdPnt_numOfEvent_,int const * const grid_to_event_, const int num_of_grids, const int max_nm_events){
    // int i = blockIdx.x;
    // int j = blockIdx.y;
    // int tid_x = threadIdx.x;
	// int tid_y = threadIdx.y;
    // blockIdx, threadIdx, blockDim are stored in a special register, copying it into a variable will waste time and space
    __shared__ float j_per_conv[ BLOCKDIM_X * BLOCKDIM_Y ];
    __shared__ int grid_to_event_i[ BLOCKDIM_X + 3 ];
    __shared__ int grid_to_event_j[ BLOCKDIM_Y + 3 ];
    __shared__ int  s_i, s_j, total_threads, max_nm_events_cached, C_matrix_index;
    
    
    int thread_id = rowMajor_2D(tid_x, tid_y, bDm_x, bDm_y);
    if(thread_id == 0){
        C_matrix_index = rowMajor_2D(i,j,num_of_grids,num_of_grids);
        C[C_matrix_index] = 0;
        s_i = grdPnt_numOfEvent_[ i ];
        s_j = grdPnt_numOfEvent_[ j ];
        total_threads = bDm_x * bDm_y;
        max_nm_events_cached = max_nm_events;
    }
    __syncthreads();
    
    if (s_i < 3 || s_j < 3) return;
    int iter_i = 0, iter_j;
    while( bDm_x * iter_i <  s_i - 2){
        
        int l = (bDm_x * iter_i + tid_x) + 1;

        if(tid_y == 0 && l <= s_i - 2)
            grid_to_event_i[tid_x + 2] = grid_to_event_[i * max_nm_events_cached +  l];
        
        if( thread_id == 0 ){
            grid_to_event_i[ 1 ] = grid_to_event_[ i * max_nm_events_cached +  bDm_x * iter_i ];
            if(iter_i){
                grid_to_event_i[0] = grid_to_event_[ i * max_nm_events_cached +  bDm_x * iter_i - 1];
            }else{
                grid_to_event_i[0] = -1;
            }
            int max_l = min(bDm_x * (iter_i + 1) + 1, s_i - 1);
            int end_event_index = max_l - bDm_x * iter_i + 1;
            grid_to_event_i[ end_event_index ] = grid_to_event_[ i * max_nm_events_cached + max_l ];
        }
        // __syncthreads();
        
            iter_j = 0;
        while( bDm_y * iter_j < s_j - 2 ){
            
            j_per_conv[ thread_id ] = 0;
            
            int m = (bDm_y * iter_j + tid_y) + 1;

            if(l <= s_i - 2 && m <= s_j - 2){
                if(tid_x == 0)
                    grid_to_event_j[tid_y + 1] = grid_to_event_[j * max_nm_events_cached +  m];

                if( thread_id == 0 ){
                    grid_to_event_j[ 0 ] = grid_to_event_[ j * max_nm_events_cached +   bDm_y * iter_j ];
                    if(s_j - 1 > bDm_y * (iter_j + 1) + 1){
                        int max_m = bDm_y * (iter_j + 1) + 1;
                        int end_event_index = bDm_y + 1 ;
                        grid_to_event_j[ end_event_index     ] = grid_to_event_[ j * max_nm_events_cached + max_m ];
                        grid_to_event_j[ end_event_index + 1 ] = grid_to_event_[ j * max_nm_events_cached + max_m + 1];
                    }else{
                        int max_m = s_j - 1;
                        int end_event_index = max_m - bDm_y * iter_j;
                        grid_to_event_j[ end_event_index     ] = grid_to_event_[ j * max_nm_events_cached + max_m ];
                        grid_to_event_j[ end_event_index + 1 ] = -1; 
                    }
                }
            }
            __syncthreads();

            if(l <= s_i - 2 && m <= s_j - 2){
                char J_is_one, J_is_half;{
                    int t_l_minus_two = grid_to_event_i[tid_x];
                    int t_l_minus_one = grid_to_event_i[tid_x + 1];
                    int t_l = grid_to_event_i[tid_x + 2];
                    int t_l_plus_one = grid_to_event_i[tid_x + 3];
                    int t_m_minus_one = grid_to_event_j[tid_y];
                    int t_m = grid_to_event_j[tid_y + 1];
                    int t_m_plus_one = grid_to_event_j[tid_y + 2];
                    int t_m_plus_two = grid_to_event_j[tid_y + 3];
    
                    char sigma_i_j_l_m, sigma_j_i_m_l_minus_one , sigma_j_i_m_plus_one_l;{
                        int two_tao = min(min(t_l_plus_one - t_l, t_l - t_l_minus_one) , min(t_m_plus_one - t_m , t_m - t_m_minus_one));
                        int time_dif = t_l - t_m;
                        sigma_i_j_l_m = (0 < time_dif) & ((2 * time_dif) <= two_tao);
                    }{
                        
                        int two_tao;
                        if (t_l_minus_two +1)
                            two_tao = min(min(t_l - t_l_minus_one, t_l_minus_one - t_l_minus_two), min(t_m_plus_one - t_m , t_m - t_m_minus_one));
                        else
                            two_tao = min(t_l - t_l_minus_one, min(t_m_plus_one - t_m , t_m - t_m_minus_one));
                        int time_dif = t_m - t_l_minus_one;
                        sigma_j_i_m_l_minus_one = (0 < time_dif) & ((2 * time_dif) <= two_tao);
                    }{
                        int two_tao;
                        if(t_m_plus_two + 1)
                            two_tao = min(min(t_l_plus_one - t_l, t_l - t_l_minus_one) , min(t_m_plus_two - t_m_plus_one, t_m_plus_one - t_m));
                        else
                            two_tao = min(min(t_l_plus_one - t_l, t_l - t_l_minus_one) , t_m_plus_one - t_m);
                        int time_dif = t_m_plus_one - t_l;
                        sigma_j_i_m_plus_one_l = (0 < time_dif) & ((2 * time_dif) <= two_tao);
                    }
                    char intermediate_condition = sigma_j_i_m_l_minus_one | sigma_j_i_m_plus_one_l;
                    J_is_one = sigma_i_j_l_m &  (!intermediate_condition);
                    J_is_half = (t_l == t_m) | (sigma_i_j_l_m & intermediate_condition);
                }
                j_per_conv[thread_id] += 0.5f * J_is_half;  
                j_per_conv[thread_id] += J_is_one;
            }
            __syncthreads();

            // for(int s = 1 << (int) ceil(log2((double)total_threads) - 1); s > WARP_SIZE ; s >>= 1){
            //     if(thread_id < s && thread_id + s < total_threads)
            //         j_per_conv[thread_id] += j_per_conv[thread_id + s];
                
            //     __syncthreads();
            // }
            // if(thread_id < WARP_SIZE)
            //     warpReduce(j_per_conv, thread_id);

            for(int s = 1 << (int) ceil(log2((double)total_threads) - 1); s > 0 ; s >>= 1){
                if(thread_id < s && thread_id + s < total_threads)
                    j_per_conv[thread_id] += j_per_conv[thread_id + s];
                
                __syncthreads();
            }


            if(thread_id == 0){
                C[C_matrix_index] += j_per_conv[0];
            }
            // __syncthreads();
                iter_j++;
        }

        // if(thread_id == 0)
            iter_i++;

        // __syncthreads();

    }
}
}
'''


import pandas as pd
import numpy as np
__all__ = [
    "createGPUFunction",
    "printTime",
    "create_grid_metadata_arrays",
    "NUMBER_OF_SECONDS_PER_HOUR"
]
NUMBER_OF_SECONDS_PER_HOUR = 3600
def create_grid_metadata_arrays(date_index: pd.DatetimeIndex, rain_event_matrix:np.ndarray):
    
    maximum_number_of_events = -1
    number_of_time_steps, number_of_event_series = rain_event_matrix.shape
    grid_to_event_map_py:list[list[int]] = []
    starting_date = date_index[0]
    grdPnt_numOfEvent:list[int] = []
    date_index_array = np.asarray(list(map(lambda x:x.total_seconds(), ((date_index - starting_date) // 3600))))
    for event_serie_index in range(number_of_event_series):
        this_event_arr:list[int] = []
        for time_step in range(number_of_time_steps):
            if rain_event_matrix[time_step, event_serie_index]:
                this_event_arr.append(date_index_array[time_step])
        grid_to_event_map_py.append(this_event_arr)
        grdPnt_numOfEvent.append(len(this_event_arr))
        maximum_number_of_events = max(maximum_number_of_events, len(this_event_arr))
    for event_serie_index in range(number_of_event_series):
        grid_to_event_map_py[event_serie_index].extend([
            -1 for _ in range(maximum_number_of_events - grdPnt_numOfEvent[event_serie_index])
        ])
    return np.asarray(grid_to_event_map_py, dtype=np.int32), np.asarray(grdPnt_numOfEvent,dtype=np.int32), maximum_number_of_events

def createGPUFunction(sourceCode:str, functionName:str, block:tuple[int,int,int]):
    from pycuda.compiler import SourceModule # type: ignore 
    compilableSrc = sourceCode.replace("BLOCKDIM_X", str(block[0])).replace("BLOCKDIM_Y", str(block[1])).replace("BLOCKDIM_Z", str(block[2]))
    Mod = SourceModule(compilableSrc, no_extern_c = True)
    return Mod.get_function(functionName)
    
def printTime(time_spent:int):
    minutes_spent = time_spent // 60
    seconds_spent = round( time_spent - 60 * minutes_spent)
    print(f"Time elapsed to run the computation :: {minutes_spent} minutes {seconds_spent} seconds ")







from numba import njit,jit, prange
import numpy as np

__all__ = [
    "ECA_Loop",
    "binomial_term_new",
    "find_p_value_new",
    "ECA_business_logic"
]

@njit(nogil= True)
def ECA_business_logic(
    event_serie_l:np.ndarray, event_serie_r:np.ndarray, 
    Delta_T : int, tau :int
):
    r_precursor = 0
    for t_l in event_serie_l:
        bin_ser_low = 0
        bin_ser_high = event_serie_r.shape[0] - 1
        while bin_ser_low <= bin_ser_high:
            bin_ser_mid =(bin_ser_low + bin_ser_high) // 2 
            t_m = event_serie_r[bin_ser_mid]
            diff = t_l - tau - t_m
            if diff > Delta_T:
                bin_ser_low = bin_ser_mid + 1
            elif diff < 0:
                bin_ser_high = bin_ser_mid - 1
            else:
                r_precursor += 1
                break
    
    if event_serie_l.shape[0] > 0:
        r_precursor /= event_serie_l.shape[0]

    r_trigger = 0
    for t_m in event_serie_r:
        bin_ser_low = 0
        bin_ser_high = event_serie_l.shape[0] - 1
        while bin_ser_low <= bin_ser_high:
            bin_ser_mid = (bin_ser_low + bin_ser_high) // 2
            t_l = event_serie_l[bin_ser_mid]
            diff = t_l - tau - t_m
            if diff > Delta_T:
                bin_ser_high = bin_ser_mid - 1
            elif diff < 0:
                bin_ser_low = bin_ser_mid + 1
            else:
                r_trigger += 1
                break    
    if event_serie_r.shape[0] > 0:
        r_trigger /= event_serie_r.shape[0]
    return r_precursor, r_trigger

# @njit((numba.int64, numba.int32[ ::1 ], numba.int32[:, ::1 ], numba.float64, numba.int64, numba.float64[:, ::1 ], numba.float64[:, ::1 ]),parallel=True, nogil= True)
@njit(parallel=True, nogil= True)
def ECA_Loop(
    number_of_grid_points: np.ndarray, grdPnt_numOfEvent:np.ndarray , 
    grid_to_event_map : np.ndarray, Delta_T : int, 
    tau :int, r_precursor_i_j : np.ndarray, 
    r_trigger_i_j : np.ndarray
):
    for i in prange(number_of_grid_points):
        # if math.floor( i * 100 / number_of_grid_points ) != math.floor( ( i - 1 ) * 100 / number_of_grid_points ):
            # print(i*100/number_of_grid_points)
        for j in np.arange(number_of_grid_points):
            numOfEvents_on_i = grdPnt_numOfEvent[i]
            numOfEvents_on_j = grdPnt_numOfEvent[j]
            r_precursor, r_trigger = ECA_business_logic(
                grid_to_event_map[i,0:numOfEvents_on_i],
                grid_to_event_map[j,0:numOfEvents_on_j],
                Delta_T, tau
            )

            r_precursor_i_j[i,j] = r_precursor
            r_trigger_i_j[i,j] = r_trigger



@njit(nogil = True)
def binomial_term_new(n:int, r:int, a:float|int, b:float|int, lgdFctr: np.ndarray) :
    data_ = lgdFctr[ n ] - lgdFctr[ n - r ] - lgdFctr[ r ] + r * np.log(a) + ( n - r ) * np.log(b)
    return np.exp(data_)

@njit(parallel = True, nogil = True)
def find_p_value_new(
    number_of_grid_points:int, grdPnt_numOfEvent:np.ndarray, 
    Delta_T:int, tau:int, 
    r_precursor_i_j:np.ndarray, r_trigger_i_j:np.ndarray, 
    pval_precursor:np.ndarray, pval_trigger:np.ndarray, 
    T:int, lgdFctr:np.ndarray
):
    prb_prec_coinc = Delta_T / (T - tau)
    for i in prange(number_of_grid_points):
        # if math.floor( i * 100 / number_of_grid_points ) != math.floor( ( i - 1 ) * 100 / number_of_grid_points ):
            # print(i*100/number_of_grid_points)
        for j in np.arange(number_of_grid_points):
            # if math.floor( j * 100 / number_of_grid_points ) != math.floor( ( j - 1 ) * 100 / number_of_grid_points ):
            #     print('\t',j*100/number_of_grid_points)
            numOfEvents_on_i = grdPnt_numOfEvent[i]
            numOfEvents_on_j = grdPnt_numOfEvent[j]
            
            prec_bnml_rgt_val = np.power(1 - prb_prec_coinc, numOfEvents_on_j)
            prec_bnml_lft_val = 1 - prec_bnml_rgt_val
            r_precursor = r_precursor_i_j[i ,j] * numOfEvents_on_i
            for K_star in np.arange(round(r_precursor), numOfEvents_on_i + 1):
                pval_precursor[i, j] += binomial_term_new( numOfEvents_on_i, K_star, prec_bnml_lft_val, prec_bnml_rgt_val, lgdFctr )


            trig_bnml_rgt_val = np.power(1 - prb_prec_coinc, numOfEvents_on_i)
            trig_bnml_lft_val = 1 - trig_bnml_rgt_val
            r_trigger = r_trigger_i_j[i ,j] * numOfEvents_on_j
            for K_star in np.arange(round(r_trigger), numOfEvents_on_j + 1):
                pval_trigger[i, j] += binomial_term_new( numOfEvents_on_j, K_star, trig_bnml_lft_val, trig_bnml_rgt_val, lgdFctr )



from numba import njit,jit, prange
import numpy as np

__all__ = [
    "ES_Loop",
    "normaliseQMatrix"
]

@njit(parallel = True, nogil = True)
def normaliseQMatrix(Q, grdPnt_numOfEvent):
    (number_of_grid_points, _) = Q.shape
    for i in prange(number_of_grid_points):
        number_of_events_i = grdPnt_numOfEvent[ i ]
        if number_of_events_i <= 2:
            continue
        Q[i, i] /= ( number_of_events_i - 2 )
        for j in np.arange( i ):
            number_of_events_j = grdPnt_numOfEvent[ j ]
            if number_of_events_j <= 2:
                continue
            normal_factor = np.sqrt( ( number_of_events_i - 2 ) * ( number_of_events_j - 2 ) )
            Q[i, j] /= normal_factor
            Q[j, i] /= normal_factor

@njit(parallel = True, nogil = True)
def ES_Loop(
    number_of_grid_points:int, grdPnt_numOfEvent:np.ndarray, 
    Tau_hlpr:np.ndarray, tauMax:float, 
    grid_to_event_map:np.ndarray, c_i_j:np.ndarray
):
    for i in prange(number_of_grid_points):
        # if math.floor( i * 100 / number_of_grid_points ) != math.floor( ( i - 1 ) * 100 / number_of_grid_points ):
        #     print(i * 100 / number_of_grid_points)
        s_i = grdPnt_numOfEvent[i]
        if s_i in [0,1,2]:
            continue
        for j in range(number_of_grid_points):
            s_j = grdPnt_numOfEvent[j]
            if s_j in [0,1,2]:
                continue
            for l in range(1, s_i - 1):
                for m in range(1, s_j - 1):
                    t_l = grid_to_event_map[i, l]
                    t_m = grid_to_event_map[j, m]
                    tao = 0.5 * min(Tau_hlpr[i, l], Tau_hlpr[j, m])
                    sigma_i_j_l_m = (t_l - t_m <= tao) and t_l > t_m
                    
                    t_l_minus_one = grid_to_event_map[i, l-1]
                    tao = 0.5 * min(Tau_hlpr[i, l-1], Tau_hlpr[j, m])
                    sigma_j_i_m_l_minus_one = t_l_minus_one < t_m and (t_m - t_l_minus_one <= tao)
                    
                    t_m_plus_one = grid_to_event_map[j, m + 1]
                    tao = 0.5 * min(Tau_hlpr[i, l], Tau_hlpr[j, m + 1])
                    sigma_j_i_m_plus_one_l = t_m_plus_one > t_l and (t_m_plus_one - t_l <= tao)
                    
                    if(sigma_i_j_l_m and not(sigma_j_i_m_l_minus_one or sigma_j_i_m_plus_one_l)):
                        c_i_j[i, j] += 1
                    elif t_l == t_m or (sigma_i_j_l_m and (sigma_j_i_m_l_minus_one or sigma_j_i_m_plus_one_l)):
                        c_i_j[i, j] += 0.5


__all__ = [
    "ECA_kernel_name",
    "P_value_kernel_name",
    "PVALUE_CODE",
    "ECA_CODE"
]

ECA_kernel_name = "ECA_kernel"
P_value_kernel_name = "pVal_kernel"

PVALUE_CODE = '''
  #define i blockIdx.x
  #define j blockIdx.y
  #define tid_x threadIdx.x
  #define tid_y threadIdx.y
  #define tid_z threadIdx.y
  #define bDm_x blockDim.x
  #define bDm_y blockDim.y
  #define bDm_z blockDim.z
  #define WARP_SIZE 32
  #define CUDART_PI_F 3.14159265f
  __forceinline__ __device__ int rowMajor_2D(int r, int c, int R, int C){
      return c + C * r;
  }

  __forceinline__ __device__ int colMajor_2D(int r, int c, int R, int C){
      return r + R * c;
  }

  extern "C"{
  __global__ void pVal_kernel( double *pVal_precursor, double *pVal_trigger, float const * const r_precursor, float const * const r_trigger, 
      const int * const grdPnt_numOfEvent_,  const int Delta_T,  const int Tau, const int num_of_grids, const int T, const double * const logedFactorial){
          
      __shared__ double answer_collection_buffer[ BLOCKDIM_X ];
      __shared__ double prb_prec_coinc_shared;
      __shared__ double prec_bnml_rgt_lg_shared;
      __shared__ double prec_bnml_lft_lg_shared;

      if( tid_x == 0 ){
          prb_prec_coinc_shared = ((double)Delta_T / (double)(T - Tau));
          prec_bnml_rgt_lg_shared = pow(1 - prb_prec_coinc_shared, grdPnt_numOfEvent_[ j ] ) ; 
          prec_bnml_lft_lg_shared = log(1 - prec_bnml_rgt_lg_shared) ;
          prec_bnml_rgt_lg_shared = log(prec_bnml_rgt_lg_shared);
          pVal_precursor[ rowMajor_2D(i, j, num_of_grids, num_of_grids) ] = 0;
      }
      __syncthreads();
      double prec_bnml_rgt_val = prec_bnml_rgt_lg_shared;
      double prec_bnml_lft_val = prec_bnml_lft_lg_shared;
      
      int K_end = grdPnt_numOfEvent_[ i ];
      int K_start = round( r_precursor[rowMajor_2D(i, j, num_of_grids, num_of_grids)] * K_end );
      int iter_i = 0;
      while( K_start + iter_i * BLOCKDIM_X < K_end ){
          answer_collection_buffer[ tid_x ] = 0;
          int K = K_start + iter_i * BLOCKDIM_X + tid_x;
          if( K <= K_end ){
              answer_collection_buffer[ tid_x ] = exp( logedFactorial[ K_end ] - logedFactorial[ K ] - logedFactorial[ K_end - K ] + K * prec_bnml_lft_val + ( K_end - K ) * prec_bnml_rgt_val );
          }

          __syncthreads();
          for(int s=1 << (int)ceil(log2((float)BLOCKDIM_X) - 1) ; s > 0 ; s >>= 1){
              if(tid_x < s && tid_x + s < BLOCKDIM_X){
                  answer_collection_buffer[ tid_x ] += answer_collection_buffer[ tid_x + s ];
              }
              __syncthreads();
          }
          if( tid_x == 0 ){
              pVal_precursor[ rowMajor_2D(i, j, num_of_grids, num_of_grids) ] += answer_collection_buffer[0];
          }
          iter_i++;
      }

      __syncthreads();

      if( tid_x == 0 ){
          prec_bnml_rgt_lg_shared = pow( 1 - prb_prec_coinc_shared, grdPnt_numOfEvent_[ i ] ) ; 
          prec_bnml_lft_lg_shared = log( 1 - prec_bnml_rgt_lg_shared ) ;
          prec_bnml_rgt_lg_shared = log( prec_bnml_rgt_lg_shared );
          pVal_trigger[ rowMajor_2D(i, j, num_of_grids, num_of_grids) ] = 0;
      }
      __syncthreads();
      prec_bnml_rgt_val = prec_bnml_rgt_lg_shared;
      prec_bnml_lft_val = prec_bnml_lft_lg_shared;
      
      K_end = grdPnt_numOfEvent_[ j ];
      K_start = round( r_trigger[rowMajor_2D(i, j, num_of_grids, num_of_grids)] * K_end );
      iter_i = 0;
      while( K_start + iter_i * BLOCKDIM_X < K_end ){
          answer_collection_buffer[ tid_x ] = 0;
          int K = K_start + iter_i * BLOCKDIM_X + tid_x;
          if( K <= K_end ){
              answer_collection_buffer[ tid_x ] = exp( logedFactorial[ K_end ] - logedFactorial[ K ] - logedFactorial[ K_end - K ] + K * prec_bnml_lft_val + ( K_end - K ) * prec_bnml_rgt_val );
          }

          __syncthreads();
          for(int s=1 << (int)ceil(log2((float)BLOCKDIM_X) - 1) ; s > 0 ; s >>= 1){
              if(tid_x < s && tid_x + s < BLOCKDIM_X){
                  answer_collection_buffer[ tid_x ] += answer_collection_buffer[ tid_x + s ];
              }
              __syncthreads();
          }
          if( tid_x == 0 ){
              pVal_trigger[ rowMajor_2D(i, j, num_of_grids, num_of_grids) ] += answer_collection_buffer[0];
          }
          iter_i++;
      }


  }
  }
'''
ECA_CODE = '''
#define i blockIdx.x
#define j blockIdx.y
#define tid_x threadIdx.x
#define tid_y threadIdx.y
#define tid_z threadIdx.y
#define bDm_x blockDim.x
#define bDm_y blockDim.y
#define bDm_z blockDim.z
#define WARP_SIZE 32

__forceinline__ __device__ int rowMajor_2D(int r, int c, int R, int C){
    return c + C * r;
}

__forceinline__ __device__ int colMajor_2D(int r, int c, int R, int C){
    return r + R * c;
}

extern "C"{
__global__ void ECA_kernel(float *r_precursor, float *r_trigger, const int * const grdPnt_numOfEvent_, const int * const grid_to_event_,  const int Delta_T, 
    const int Tau, const int num_of_grids, const int max_nm_events){
    __shared__ int  matrix_interm_buffer[ BLOCKDIM_X ];
    //matrix to temp store results (l/m) in a column major order
    __shared__ int outer_event_time_buffer[ BLOCKDIM_X ];
    __shared__ int inner_event_time_buffer[ BLOCKDIM_X ];
    __shared__ int s_i, s_j,  r_matrix_index;
    if( tid_x == 0 ){ 
        r_matrix_index = rowMajor_2D(i, j, num_of_grids, num_of_grids);
        r_precursor[ r_matrix_index ] = 0;
        r_trigger[ r_matrix_index ] = 0; 
        s_i = grdPnt_numOfEvent_[ i ];
        s_j = grdPnt_numOfEvent_[ j ];
    }
        
    __syncthreads();
    if (s_i < 1 || s_j < 1) return;
    int iter_i = 0, iter_j;
    
    while( iter_i * BLOCKDIM_X < s_i){
        
        int l = iter_i * BLOCKDIM_X + tid_x;
        
        if( l < s_i){
            outer_event_time_buffer[ tid_x ] = grid_to_event_[ i * max_nm_events + l];
        }
        
        matrix_interm_buffer[ tid_x ] = 0;

        // if( tid_x == 0 ){
            iter_j = 0;
        // } 

        // __syncthreads();

        while(iter_j * BLOCKDIM_X < s_j){
            
            int m = iter_j * BLOCKDIM_X + tid_x;
            
            if( m < s_j ){
                inner_event_time_buffer[ tid_x ] = grid_to_event_[ j * max_nm_events + m ];
            }

            int max_m = min( (iter_j + 1)  * BLOCKDIM_X, s_j) - ( iter_j * BLOCKDIM_X ) - 1;
            int bin_ser_low = 0, bin_ser_high = max_m;
            m = (bin_ser_low + bin_ser_high) / 2;
            
            __syncthreads();
            if( l < s_i ){
                for( int bin_ser_iter = 1 ; bin_ser_iter <= BLOCKDIM_X ; bin_ser_iter <<= 1){
                    int diff = outer_event_time_buffer[ tid_x ] - Tau - inner_event_time_buffer[ min( max(m, 0), max_m ) ];
                    char undershot = diff < 0 ;
                    char overshot =  diff > Delta_T ;
                    char perfect = (undershot | overshot) ^ 1 ;
                    bin_ser_low = ( overshot * ( m + 1 ) ) + ( ( overshot ^ 1 ) * bin_ser_low);
                    bin_ser_high = ( undershot * ( m - 1 ) ) + ( ( undershot ^ 1 ) * bin_ser_high );
                    m = (bin_ser_low + bin_ser_high) / 2;
                    matrix_interm_buffer[ tid_x ] |= perfect;
                }
            }
            __syncthreads();

            // if( tid_x == 0 ){
                iter_j ++;
            // }

        
        }

        for(int s=1 << (int)ceil(log2((float)BLOCKDIM_X) - 1) ; s > 0 ; s >>= 1){
            if(tid_x < s && tid_x + s < BLOCKDIM_X){
                matrix_interm_buffer[ tid_x ] += matrix_interm_buffer[ tid_x + s ];
            }
            __syncthreads();
        }

        if( tid_x == 0 ){
            r_precursor[ r_matrix_index ] += matrix_interm_buffer[ 0 ];
        }
            iter_i++;
        // __syncthreads();
    
    }

    if(tid_x == 0){
        r_precursor[ r_matrix_index ] /= s_i;
    }

    // __syncthreads();

        iter_i = 0;

    while( iter_i * BLOCKDIM_X < s_j ){
        
        int m = iter_i * BLOCKDIM_X + tid_x;
        
        if( m < s_j ){
            outer_event_time_buffer[ tid_x ] = grid_to_event_[ j * max_nm_events + m];
        }
        
        matrix_interm_buffer[ tid_x ] = 0;

        // if( tid_x == 0 ){
            iter_j = 0;
        // } 

        // __syncthreads();

        while(iter_j * BLOCKDIM_X < s_i){
            
            int l = iter_j * BLOCKDIM_X + tid_x;
            
            if( l < s_i ){
                inner_event_time_buffer[ tid_x ] = grid_to_event_[ i * max_nm_events + l ];
            }

            int max_l = min( (iter_j + 1)  * BLOCKDIM_X, s_i) - ( iter_j * BLOCKDIM_X ) - 1;
            int bin_ser_low = 0, bin_ser_high = max_l;
            l = (bin_ser_low + bin_ser_high) / 2;
            
            __syncthreads();
            if( m < s_j ){
                for( int bin_ser_iter = 1 ; bin_ser_iter <= BLOCKDIM_X ; bin_ser_iter <<= 1){
                    int diff = inner_event_time_buffer[ min( max(l, 0), max_l ) ] - Tau - outer_event_time_buffer[ tid_x ] ;
                    char undershot = diff < 0 ;
                    char overshot =  diff > Delta_T ;
                    char perfect = (undershot | overshot) ^ 1 ;
                    bin_ser_high = ( overshot * ( l - 1 ) ) + ( ( overshot ^ 1 ) * bin_ser_high);
                    bin_ser_low = ( undershot * ( l + 1 ) ) + ( ( undershot ^ 1 ) * bin_ser_low );
                    l = (bin_ser_low + bin_ser_high) / 2;
                    matrix_interm_buffer[ tid_x ] |= perfect;
                }
            }
            __syncthreads();

            // if( tid_x == 0 ){
                iter_j ++;
            // }

        
        }

        for(int s=1 << (int)ceil(log2((float)BLOCKDIM_X) - 1) ; s > 0 ; s >>= 1){
            if(tid_x < s && tid_x + s < BLOCKDIM_X){
                matrix_interm_buffer[ tid_x ] += matrix_interm_buffer[ tid_x + s ];
            }
            __syncthreads();
        }

        if( tid_x == 0 ){
            r_trigger[ r_matrix_index ] += matrix_interm_buffer[ 0 ];
        }
        // __syncthreads();
            iter_i++;
    
    }

    if(tid_x == 0){
        r_trigger[ r_matrix_index ] /= s_j;
    }

    

}
}
'''
